// Retro Avatars
// -------------

// By Richard Phipps
// (Open source, but a credit if used would be nice!)

#include <allegro.h>
#include <stdio.h>
#include <string.h>

// Maximum number of sprites to grow at once.
#define MAX_GROW           1

BITMAP *src, *buffer, *fnt, *dest;
BITMAP *grow[MAX_GROW];
PALETTE pal;
char title[16]; // Used for typed text.

// Some global variables.
int len, col, x_sym, y_sym, noise;
unsigned long code;

void destroy_grow_bmps(void);
void setup_grow_bmps(void);
void grow_all_bmps(int r);
void grow_bmp(int g, int r);
void copy_part_from_grow(int g);
void fill_part(int c);
void draw_grow_bmp(int g);
void draw_all_grow_bmps(void);
void draw_text(BITMAP *bmp, int x, int y, char *text, int scale);
void mirror(int g);
int check_for_filled_edge(int g);
void enhance_face(int g);
void outline_area(int g, int col);
void remove_noise(int g, int type);
void trim_area(int g, int col, int x2, int y1);

void draw_zoomed_image(BITMAP *bmp, BITMAP *dest, int x1, int y1, int x2, int y2, int scale, int col);
     
int main(void)
{
   char name[4096]; // Used for file path.
   int x, y, g, c, ky;
   
   // you should always do this at the start of Allegro programs
   allegro_init();

   // set up the keyboard handler
   install_keyboard(); 
   
   set_color_depth(32);
   
   // set a graphics mode sized 292 x 256 in a window
   if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, 292, 256, 0, 0) != 0) {
      set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
      allegro_message("Unable to set any graphic mode\n%s\n", allegro_error);
      return 0;
   }
   install_mouse();
   
   sprintf(name, ""); // Blank file path.
  
   buffer = create_bitmap(292, 256); // Create double buffer for screen.
   clear(buffer); // Clear of any previous memory data.
   
   src = create_bitmap_ex(8, 12, 12); // Source for sprite area (only used now for size of sprite).
   clear(src);
   
   dest = create_bitmap_ex(32, 112, 112); // Destination buffer for saved image.
   clear(dest);
   
   fnt = load_bitmap("font.bmp", pal); // Load font gfx and palette.
   if (!fnt)
   {     
     set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
     allegro_message("Unable to load font gfx");
     return 0;        
   }

   set_palette(pal);
   
   start:
   clear_keybuf();
   clear(buffer);
   
   srand(0); // Reset random number seed.
   setup_grow_bmps();
   
   // Setup some default global values.
   len = 0;
   x_sym = 1;
   y_sym = 0;
   noise = 4;
   
   // Blank any old typed text.
   for (g = 0 ; g < 12 ; g++)
   {
    title[g] = 32;    
   }

// For testing only..
//goto end;

   // Draw title text (x4) zoom.
   draw_text(buffer, 84, 4, "RETRO", 4);
   draw_text(buffer, 60, 28, "AVATARS", 4);
   
   draw_text(buffer, 24, 100, "ENTER NAME", 4);
   blit(buffer, screen, 0, 0, 0, 0, buffer->w, buffer->h); // Blit double buffer to the screen.
   
   // Let user enter typed text.
   c = 0;
   do
   {
    ky = readkey() & 0xff; // Wait until a key is pressed and then find ASCII value.
    if (ky >= 97 && ky <= 122) ky = ky - 32; // lower case keys, so make upper case.
    
    if ((ky == 32 || (ky >= 65 && ky <= 90) || (ky >=48 && ky <= 57)) && c < 9) // Space, letters, numbers and less than 9 space in length.
    {
     title[c] = ky;
     c++;
    }
    
    if (ky == 8 && c > 0) // Backspace
    {
     title[c - 1] = 32;
     c--;       
    }

    draw_text(buffer, 36, 128, title, 4); // Update typed text
    blit(buffer, screen, 0, 0, 0, 0, buffer->w, buffer->h);
                     
   } while (ky != 27 && ky != 13); // Repeat until Escape or Return is pressed.
   if (ky == 27) goto exit; // Escape, so exit.
   
   
   code = 0; 
   for (g = 0 ; g < 32 ; g++)
   {
    code = code + (title[g % c] * rand() % 1024); // Create semi-unique code for random generator from text.
   }
   end:
   //code = 0;
   
   srand(code); // Seed random generator with code.
   col = makecol((rand() % 192) + 64, (rand() % 192) + 64, (rand() % 192 + 64)); // Random (but not too dark) colour.
        
   len = 9 - c;
   clear(buffer); // Clear 'screen'
   grow_all_bmps(0); // Generate sprite
   
   // This section is commented out, and used for showing a LOT of images at once for testing.
   /*
   for (y = 0 ; y < 8 ; y++)
   {
   for (x = 0 ; x < 10 ; x++)
   {
    srand(code);
    col = makecol((rand() % 192) + 64, (rand() % 192) + 64, (rand() % 192 + 64));
   grow_all_bmps(0);   
    draw_zoomed_image(grow[0], buffer, x * 28, y * 28, 0, 0, 2, col);
    code++; 
   }
}  
*/
   // Show generated sprite on screen, as well as text underneath.
   draw_zoomed_image(grow[0], buffer, 48, 8, 0, 0, 16, col); 
   draw_text(buffer, 40 + (len * 12), 208, title, 4);
   
   draw_text(buffer, 0, 232, "PRESS S TO SAVE AVATAR", 1);
   draw_text(buffer, 0, 240, "PRESS N TO ENTER A NEW NAME", 1);
   draw_text(buffer, 0, 248, "PRESS ESCAPE TO EXIT", 1);
   
   blit(buffer, screen, 0, 0, 0, 0, buffer->w, buffer->h); // Double buffer to the screeb.
      
   do
   {
                       
    if (key[KEY_S]) // Save image
    {
     clear(dest); // Clear dest image.
     draw_zoomed_image(grow[0], dest, 14, 4, 0, 0, 7, col); // Draw image to dest buffer
     draw_text(dest, 4 + (len * 6), 94, title, 2); // And text underneath
        
     sprintf(name, "%s.bmp", title); // Add .bmp extender
                     
     save_bitmap(name, dest, NULL); // and save, easy!
     
     goto start; // Repeat as if N had been pressed.
    }
    
    if (key[KEY_N]) goto start; // New image
    
    blit(buffer, screen, 0, 0, 0, 0, buffer->w, buffer->h); // Keep updating the double buffer
    rest(1); // Give back some CPU time to the OS.
    
   } while (!key[KEY_ESC]);

 exit:
     
 // Free memory.
 destroy_grow_bmps();
 destroy_bitmap(dest);
 destroy_bitmap(src);
 destroy_bitmap(fnt);
 destroy_bitmap(buffer);
 
 return 0;
}
END_OF_MAIN();


void setup_grow_bmps(void)
{
 int g;
 
 // Create 8-bit images for each 'grow' sprite (using size from src buffer).
 for (g = 0 ; g < MAX_GROW ; g++)
 {
  grow[g] = create_bitmap_ex(8, src->w, src->h);
  clear(grow[g]);
 }
}

void destroy_grow_bmps(void)
{
 int g;
 
 for (g = 0 ; g < MAX_GROW ; g++)
 {
  destroy_bitmap(grow[g]);
 }     
}

// Used to draw a zoomed in pixel version of a sprite.
void draw_zoomed_image(BITMAP *bmp, BITMAP *dest, int x1, int y1, int x2, int y2, int scale, int col)
{
 int x, y, c;
 int sx = scale;
 int sy = scale;

 for (y = 0 ; y < bmp->h ; y++)
 {
  for (x = 0 ; x < bmp->w ; x++)
  {
   c = getpixel(bmp, x, y); // Read pixel from source image.
   
   // Draw all in one colour or with font palette (for testing purposes to see eye, nose & mouth areas).
   if (c >= 255)
   {
    // And draw coloured square for each 'lit' pixel.
    rectfill(dest, x1 + (sx * x), y1 + (sy * y), (x1 + (sx * x) + sx) - 1, (y1 + (sy * y) + sy) - 1, col);
    //makecol(pal[c].r * 4, pal[c].g * 4, pal[c].b * 4));      
   }
  }    
 }
}

void grow_all_bmps(int r)
{
 int g;
 
 for (g = 0 ; g < MAX_GROW ; g++)
 {
  grow_bmp(g, r); // Grow each sprite in list.    
 }     
}

void grow_bmp(int g, int r)
{
 int x, y, c, c2;
 
 for (y = 0 ; y < src->h ; y++)
 {
  for (x = 0 ; x < src->w ; x++)
  {
   // Very simple! The higher the value of C the more solid pixels are placed.
   c = 158;
   
   if (rand() % 356 > c)
   {
    putpixel(grow[g], x, y, 0); // Empty          
   }
   else
   {
    putpixel(grow[g], x, y, 255); // Solid
   }     
  }
 }
 // All other colour values are used for reserved area of the image (eyes, nose & mouth).
 
 
 // Remove single pixel isolated noise and join up gaps 
 // (The higher the global noise value, the more times this is done).
 remove_noise(g, 0);
  
 mirror(g); // Make processed image symmetrical.
  
 enhance_face(g); // Identify (or create), and enhance facial characteristics.
}

void enhance_face(int g)
{
 int x, y, ny, c, a;
 int eyes = 0; // Found eyes, nose, mouth. 0 = no, 1 = yes.
 int nose = 0;
 int mouth = 0;
 int hx = (src->w / 2) - 1; // Half width of sprite variable.
 
 // Detect eyes one pixel away from horizontal centre (look from just below the top edge to the middle of the vertical height).
 for (y = 1 ; y < src->h / 2 ; y++)
 {
  if (getpixel(grow[g], hx - 1, y) == 0) // 0 - Empty, Pixel?
  {
   floodfill(grow[g], hx - 1, y, 1); // Mark area with reserved colour 1 (normal colours are 0 - empty & 255 - solid).
   if (check_for_filled_edge(g) == 0) break; // If this eye area doesn't touch the edges of the image, then stop searching.
   floodfill(grow[g], hx - 1, y, 0); // It reaches the edge, so refill as 0 - empty and keep looking.
  }
  if (getpixel(grow[g], hx - 2, y) == 0) // Any potential eye areas one pixel further away?
  {
   floodfill(grow[g], hx - 2, y, 1);
   if (check_for_filled_edge(g) == 0) break;
   floodfill(grow[g], hx - 2, y, 0);
  }
 }
 
 if (y == src->h / 2) // Ok, we didn't find anything!
 {
 
 // Try to make eyes from any centre pixels (converting them to one pixel further away. i.e. xx -> x  x
 for (y = 1 ; y < src->h ; y++)
 {
  if (getpixel(grow[g], hx - 1, y) == 255 && getpixel(grow[g], hx, y) == 0)
  {
    putpixel(grow[g], hx - 1, y, 0);                     
    putpixel(grow[g], hx, y, 255);
    putpixel(grow[g], hx - 1, y + 1, 0); // Make the eye 2 pixels (at least) high.            
    
    
   floodfill(grow[g], hx - 1, y, 1);
   if (check_for_filled_edge(g) == 0) break;
   floodfill(grow[g], hx - 1, y, 0);                        
  }
 }
}
 
 ny = y + 1;
 if (y < src->h) eyes = 1; // Ok, we did find eyes
 
 if (eyes == 0) // Still NO eyes.
 {
 // Ok, create fake eyes!
  y = 1 + rand() % (src->h / 2);
  
  putpixel(grow[g], hx - 1, y, 1);
  outline_area(g, 1); // Outline to protect area.
  
  eyes = 1;
  ny = y + 1;
 }
 

 // Remove any joined up eyes (i.e xx instead of x  x)
 for (y = 1 ; y < src->h ; y++)
 {
  if (getpixel(grow[g], hx, y) == 1)
  {
   putpixel(grow[g], hx, y, 255);
   putpixel(grow[g], hx - 1, y, 1);
  }
  else
  {   
   if (getpixel(grow[g], hx - 2, y) == 1) putpixel(grow[g], hx, y, 255);     
   if (getpixel(grow[g], hx - 1, y) == 1) putpixel(grow[g], hx, y, 255);  
  }
 }

 mirror(g); // Miror all eye work.

 if (eyes == 1) outline_area(g, 1); // Protect eyes with solid outline.
 
 // -------
 
 // Detect nose
 for (y = ny ; y < src->h ; y++)
 {
  if (getpixel(grow[g], hx, y) == 0)
  {
   floodfill(grow[g], hx, y, 2); // Fill with area colour 2.
   if (check_for_filled_edge(g) == 0) break;
   floodfill(grow[g], hx, y, 0);
  }
 }
 if (y < 10) nose = 1;

 // No nose?
 if (nose == 0)
 {
  // Ok, we won't find a mouth either, but we need to make a nose/mouth one out of any open sections (regardless of touching the edge)
  for (y = ny ; y < src->h - 1 ; y++)
  {
   if (getpixel(grow[g], hx, y) == 0)
   {
    putpixel(grow[g], hx, y, 2);
    nose = 1;

    goto skip; // Found a nose in the centre.
   }
  }
  // Try to find a nose/mouth one pixel away which we can join up. i.e. x  x -> xxxx
  for (y = ny ; y < src->h - 1 ; y++)
  {
   if (getpixel(grow[g], hx - 1, y) == 0)
   {
    putpixel(grow[g], hx - 1, y, 2);                     
    putpixel(grow[g], hx, y, 2);
    nose = 1;
    goto skip;
   }
  }

  
  // Ok, NOTHING, just create fake mouth/nose!
  y = ny + 1 + rand() % (src->h / 3);
  if (y > src->h - 2) y = src->h - 2;
  
  putpixel(grow[g], hx, y, 2);
  nose = 1;
  ny = y + 1;
  goto skip; // No need to check for mouth as in same X position as a nose.
 }
 
 ny = y + 1;
 
 // --------
 // Detect mouth
 for (y = ny ; y < src->h ; y++)
 {
  if (getpixel(grow[g], hx, y) == 0)
  {
   floodfill(grow[g], hx, y, 3);
   if (check_for_filled_edge(g) == 0) break;
   floodfill(grow[g], hx, y, 0);
  }
 }
 if (y < src->h) mouth = 1;
 
 if (mouth == 0) // Still no mouse, so look one pixel further away and then if found, join up.
 {
  for (y = ny ; y < src->h - 1 ; y++)
  {
   if (getpixel(grow[g], hx - 1, y) == 0)
   {
    putpixel(grow[g], hx, y, 0);
    floodfill(grow[g], hx, y, 3);
    if (check_for_filled_edge(g) == 0) break;
    floodfill(grow[g], hx, y, 0);

   }
  }         
 }
 if (y < src->h) mouth = 1;
 
 skip:
     
 
 // Outline mouth / nose to protect and stop surrounding gfx 'bleeding' in.
 if (mouth == 1) outline_area(g, 3); // Mouth
 if (nose == 1) outline_area(g, 2); // Nose
 

 if (eyes == 1) trim_area(g, 1, 3, 3); // Trim eyes to no more than 3 x 3                      
 if (nose == 1 && mouth == 1) trim_area(g, 2, 3, 3); // Trim nose to no more than 3 x 3. Mouth can be bigger..;
 mirror(g); // Mirror to fix changes symmetrically.
   
 // Now search for any fill in any holes that doesn't leak to the edge of the sprite (passing over eyes, mouth and nose areas).
 for (y = 1 ; y < src->h - 1 ; y++)
 {
  for (x = 1 ; x < hx - 1 ; x++)
  {
   if (getpixel(grow[g], x, y) == 0)
   {      
    floodfill(grow[g], x, y, 4);
    c = check_for_filled_edge(g);
    
    if (c != 0)
    {
      floodfill(grow[g], x, y, 0);                    
    }
    else
    {
     floodfill(grow[g], x, y, 255);                  
    }           
   }
  }
 }
        
 mirror(g); // Mirror finally (neccessary?)
}

int check_for_filled_edge(int g)
{
 int x, y, c;
 
 for (y = 0 ; y < src->h ; y++)
 {
  c = getpixel(grow[g], 0, y);
  if (c > 0 && c < 255) return 1;    
  
  c = getpixel(grow[g], src->w - 1, y);
  if (c > 0 && c < 255) return 1;    
 }
 
 for (x = 0 ; x < src->w ; x++)
 {
  c = getpixel(grow[g], x, 0);
  if (c > 0 && c < 255) return 2;    
  
  c = getpixel(grow[g], x, src->h - 1);
  if (c > 0 && c < 255) return 2;    
 }
 
 return 0;
}

void outline_area(int g, int col)
{
 int x, y, c;
 
 for (y = 0 ; y < src->h ; y++)
 {
  for (x = 0 ; x < src->w ; x++)
  {
   c = getpixel(grow[g], x, y);
   
   if (c == col)
   {
    // diagonals are only outlined if blank (and not another reserved area).
    if (getpixel(grow[g], x - 1, y - 1) == 0) putpixel(grow[g], x - 1, y - 1, 255);  
    if (getpixel(grow[g], x, y - 1) != col) putpixel(grow[g], x, y - 1, 255);
    if (getpixel(grow[g], x + 1, y - 1) == 0) putpixel(grow[g], x + 1, y - 1, 255);
    if (getpixel(grow[g], x - 1, y) != col) putpixel(grow[g], x - 1, y, 255);
    if (getpixel(grow[g], x + 1, y) != col) putpixel(grow[g], x + 1, y, 255);
    if (getpixel(grow[g], x - 1, y + 1) == 0) putpixel(grow[g], x - 1, y + 1, 255);
    if (getpixel(grow[g], x, y + 1) != col) putpixel(grow[g], x, y + 1, 255);
    if (getpixel(grow[g], x + 1, y + 1) == 0) putpixel(grow[g], x + 1, y + 1, 255);
   }     
  }
 }      
}

void trim_area(int g, int col, int x2, int y2)
{
 int x, y, nx, ny, rx, c;
 
 nx = -1;
 ny = -1;
 for (y = 0 ; y < src->h ; y++)
 {
  for (x = 0 ; x < src->w / 2 ; x++)
  {
   if (col == 1) rx = x;
   if (col > 1) rx = ((src->w / 2) - 1) - x;
   
   c = getpixel(grow[g], rx, y);
    
   if (c == col)
   {
    if (nx == -1) nx = x;
    if (ny == -1) ny = y;
    
    if (x >= nx + x2) putpixel(grow[g], rx, y, 255);
    if (y >= ny + y2) putpixel(grow[g], rx, y, 255); 
   }     
  }
 }         
}

void remove_noise(int g, int type)
{
 int x, y, c, c2;
 
for (c2 = 0 ; c2 < noise ; c2++)
 { 

 // Remove isolated pixels
 for (y = 0 ; y < src->h ; y++)
 {
  for (x = 0 ; x < src->w ; x++)
  {
   c = getpixel(grow[g], x, y);
  
   if (c == 0) // Black pixel.
   {
    c = getpixel(grow[g], x, y - 1) + getpixel(grow[g], x - 1, y) + getpixel(grow[g], x - 1, y - 1) + 
    getpixel(grow[g], x + 1, y - 1) +
    getpixel(grow[g], x + 1, y) +  getpixel(grow[g], x, y + 1) + getpixel(grow[g], x - 1, y + 1) + 
    getpixel(grow[g], x + 1, y + 1);
    
    if (type == 0 && c >= 8 * 255) putpixel(grow[g], x, y, 255);
    if (type == 1 && c >= 7 * 255) putpixel(grow[g], x, y, 255);
    
    
    // Join up 'one pixel' horizontal and vertical gaps, (adds a little order to the image).
    if (getpixel(grow[g], x, y + 1) == 255 && getpixel(grow[g], x, y - 1) == 255 && getpixel(grow[g], x - 1, y) != 255
    && getpixel(grow[g], x + 1, y) != 255 && rand() % 5 > 2) putpixel(grow[g], x, y, 255);

    if (getpixel(grow[g], x, y + 1) != 255 && getpixel(grow[g], x, y - 1) != 255 && getpixel(grow[g], x - 1, y) == 255
    && getpixel(grow[g], x + 1, y) == 255 && rand() % 5 > 2) putpixel(grow[g], x, y, 255);
    
    
   }    
  }
}

  // Remove isolated pixels
 for (y = 0 ; y < src->h ; y++)
 {
  for (x = 0 ; x < src->w ; x++)
  {
   c = getpixel(grow[g], x, y);
     
   if (c == 255) // Solid coloured pixel.
   {
    // Add up surrounding pixels.
    c = getpixel(grow[g], x, y - 1) + getpixel(grow[g], x - 1, y) + getpixel(grow[g], x - 1, y - 1) + 
    getpixel(grow[g], x + 1, y - 1) +
    getpixel(grow[g], x + 1, y) +  getpixel(grow[g], x, y + 1) + getpixel(grow[g], x - 1, y + 1) + 
    getpixel(grow[g], x + 1, y + 1);
    
    // No lit pixels around this one.
    if (c <= 0) putpixel(grow[g], x, y, 0);
   }
}
}
}     
}

void mirror(int g)
{
 int x, y, c;
 
   for (y = 0 ; y < src->h ; y++)
  {
   // X Mirror
   for (x = 0 ; x < src->w / 2 ; x++)
   {
    c = getpixel(grow[g], x, y);

    if (x_sym == 1)
    {
     putpixel(grow[g], grow[g]->w - 1 - x, y, c);
    }
   }
  }
  
  for (y = 0 ; y < src->h / 2 ; y++)
  {
   // Y Mirror
   for (x = 0 ; x < src->w ; x++)
   {
    c = getpixel(grow[g], x, y);

    if (y_sym == 1)
    {
     putpixel(grow[g], x, grow[g]->h - 1 - y, c);
    }
   }
  }    
}

void draw_text(BITMAP *bmp, int x, int y, char *text, int scale)
{
 int a, c;
 
 for (a = 0 ; a < strlen(text) ; a++)
 {
  c = text[a];
  
  if (c == 32) // Space
  {
   stretch_blit(fnt, bmp, 156, 0, 6, 6, x, y, 6 * scale, 6 * scale);  
   x = x + (6 * scale);
   continue;      
  }
  if (c >= 65 && c <= 90) // Letters
  {
   stretch_blit(fnt, bmp, (c - 65) * 6, 0, 6, 6, x, y, 6 * scale, 6 * scale);     
  }
  
  if (c >= 48 && c <= 57) // Numbers
  {
   stretch_blit(fnt, bmp, (c - 21) * 6, 0, 6, 6, x, y, 6 * scale, 6 * scale);     
  }
  
  x = x + (6 * scale);
 }     
}
